/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-  */
/* choonvdeliver.c
 *
 * Copyright (C) 2019  Giam Teck Choon <choon@choon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <bglibs/str.h>
#include <bglibs/iobuf.h>
#include "choonvdeliver.h"

/*
 * Global variables
 */
char *ufline = NULL;
char *rpline = NULL;
char *dtline = NULL;
char *home = NULL;
char *user = NULL;
char *ext = NULL;
char *host = NULL;
char *sender = NULL;
char *recipient = NULL;
char *vuser = NULL;
char *maildir = NULL;
char *email = NULL;

str hostname = {0,0,0};
str msgname = {0,0,0};
str tmpstr = {0,0,0};
str tmp2str = {0,0,0};
str tmp3str = {0,0,0};

uintmax_t msgsize = 0;
struct stat st;

/*
 * Local variables
 */
static vpwentry vpw;
static int run_under_vdeliver = 0;
static int do_deliver = 0;

static str nstr = {0,0,0};
static str tstr = {0,0,0};

void freemem(void)
{
	if (hostname.s) str_free(&hostname);
	if (msgname.s) str_free(&msgname);
	if (tmpstr.s) str_free(&tmpstr);
	if (tmp2str.s) str_free(&tmp2str);
	if (tmp3str.s) str_free(&tmp3str);
}

static void failfree(void)
{
	freemem();
	if (nstr.s) str_free(&nstr);
	if (tstr.s) str_free(&tstr);
}

/*
 * Returns zero on success, or -1 on error, with errno set to indicate
 * the cause of the error.
 */
static int env_s(char *name, const char *value)
{
	if (!getenv(name)) {
		return setenv(name, value, 1); /* <stdlib.h> */
	}
	return 0;
}

/* https://github.com/bruceg/vmailmgr/blob/master/lib/vpwentry/vpwentry.h */
static void vpw_export_env(vpwentry *vpw)
{
	if (getenv("VUSER") && getenv("MAILDIR")) {
		run_under_vdeliver = 1;
	}

	env_s("VUSER", vpw->name.s);
	vuser = getenv("VUSER");

	env_s("MAILDIR", vpw->directory.s);
	maildir = getenv("MAILDIR");
	if (maildir && is_maildir(maildir)) {
		if (vpw->has_mailbox && vpw->is_mailbox_enabled) {
			do_deliver = 1;
		}
	}
	CATCH_NE_ONE(str_copy3s(&tmpstr, vuser, "@", host),
		     exit_cb(FAIL_SOFT, failfree));
	env_s("VUSER_EMAIL", tmpstr.s);
	email = getenv("VUSER_EMAIL");

	if (!hostname.s) {
		CATCH_NE_ZERO(str_hostname(&hostname),
			      exit_cb(FAIL_SOFT, failfree));
		env_s("HOSTNAME", hostname.s);
	}

	if (!msgname.s) {
		CATCH_NE_ZERO(maildirformatname(&msgname, hostname.s),
			      exit_cb(FAIL_SOFT, failfree));
	}
}

/*
 * Remove mail found in tmp folder.
 */
void delivery_fail(void)
{
	if (do_deliver && msgname.s) {
		str_truncate(&tstr, 0);
		CATCH_NE_ONE(str_copy3s(&tstr, maildir, "/tmp/", msgname.s),
			     exit_cb(FAIL_SOFT, failfree));
		if (is_file(tstr.s)) {
			CATCH_NE_ZERO(unlink(tstr.s),
				      exit_cb(FAIL_SOFT, failfree));
		}
		str_truncate(&tstr, 0);
		CATCH_NE_ONE(str_copy5s(&tstr, maildir, "/.", SPAMDIRNAME,
					"/tmp/", msgname.s),
			     exit_cb(FAIL_SOFT, failfree));
		if (is_file(tstr.s)) {
			CATCH_NE_ZERO(unlink(tstr.s),
				      exit_cb(FAIL_SOFT, failfree));
		}
	}
}

/*
 * Move mail found in tmp to its respective new folder.
 */
void delivery_final(void)
{
	if (do_deliver && msgname.s) {
		str_truncate(&tstr, 0);
		CATCH_NE_ONE(str_copy3s(&tstr, maildir, "/tmp/", msgname.s),
			     exit_cb(FAIL_SOFT, failfree));
		str_truncate(&nstr, 0);
		CATCH_NE_ONE(str_copy3s(&nstr, maildir, "/new/", msgname.s),
			     exit_cb(FAIL_SOFT, failfree));
		if (is_file(tstr.s)) {
			CATCH_NE_ZERO(link(tstr.s, nstr.s),
				      exit_cb(FAIL_SOFT, failfree));
			CATCH_NE_ZERO(unlink(tstr.s),
				      exit_cb(FAIL_SOFT, failfree));
		}
		str_truncate(&tstr, 0);
		CATCH_NE_ONE(str_copy5s(&tstr, maildir, "/.", SPAMDIRNAME,
					"/tmp/", msgname.s),
			     exit_cb(FAIL_SOFT, failfree));
		str_truncate(&nstr, 0);
		CATCH_NE_ONE(str_copy5s(&nstr, maildir, "/.", SPAMDIRNAME,
					"/new/", msgname.s),
			     exit_cb(FAIL_SOFT, failfree));
		if (is_file(tstr.s)) {
			CATCH_NE_ZERO(link(tstr.s, nstr.s),
				      exit_cb(FAIL_SOFT, failfree));
			CATCH_NE_ZERO(unlink(tstr.s),
				      exit_cb(FAIL_SOFT, failfree));
		}
	}
}

int main(void)
{
#define ENV_VAR(VAR,ENV) VAR = getenv(#ENV); if(!VAR) { \
	errmsg(#ENV " is not set\n"); exit(FAIL_HARD); }
	ENV_VAR(home, HOME);
	ENV_VAR(user, USER);
	ENV_VAR(ext, EXT);
	ENV_VAR(host, HOST);
	ENV_VAR(recipient, RECIPIENT);
	ENV_VAR(ufline, UFLINE);
	ENV_VAR(rpline, RPLINE);
	ENV_VAR(dtline, DTLINE);
#undef ENV_VAR
	sender = getenv("SENDER");
	if (!sender) {
		sender = getenv("NEWSENDER");
	}

	if (!is_dir(home)) {
		errmsg("homedir not exists\n");
		failfree();
		return FAIL_SOFT;
	}
	if (chdir(home)) {
		errmsg("Unable to chdir to homedir\n");
		failfree();
		return FAIL_SOFT;
	}
	if (vpwlookup(&vpw, ext, "./passwd.cdb") != 1) {
		if (vpwlookup(&vpw, "+", "./passwd.cdb") != 1) {
			errmsgf("{Invalid or unknown virtual user '}s{'\n}",
				ext);
			failfree();
			return FAIL_HARD;
		}
	}
	if (vpw.expiry < (unsigned)time(0)) {
		errmsgf("{Virtual user '}s{' has expired\n}", ext);
		failfree();
		return FAIL_HARD;
	}
	vpw_export_env(&vpw);

	if (make_seekable(STDIN_FILENO) != STDIN_FILENO) {
		errmsg("make_seekable STDIN_FILENO failed.\n");
		failfree();
		return FAIL_SOFT;
	}

	if (lseek(STDIN_FILENO, 0L, SEEK_SET) != 0) {
		errmsgf("{lseek STDIN_FILENO failed. Line='}d{'\n}", __LINE__);
		failfree();
		return FAIL_SOFT;
	}

	if (fstat(STDIN_FILENO, &st) == -1) {
		errmsg("Failed to stat STDIN message size\n");
		failfree();
		return FAIL_SOFT;
	}
	if (st.st_blocks * 512 < st.st_size) {
		msgsize = st.st_size;
	} else {
		msgsize = st.st_blocks * 512;
	}

	if (do_deliver) {
		check_quota(&vpw);
		if (check_spam()) {
			delivery_fail();
			failfree();
			if (sender) {
				outmsgf("{Message from '}s{' to '}s{' detected"
					" as Spam.\n}", sender, email);
			} else {
				outmsgf("{Message from 'null' to '}s{'"
					" detected as Spam.\n}", email);
			}
			return SUCCESS_STOP;
		}
	}
	/*
	 * Forwarders 
	 */
	if (sender && vpw.forwards.len > 0 && vpw.forwards.s[0] != '\0') {
		str to = {0,0,0};
		str_copyb(&to, vpw.forwards.s, vpw.forwards.len);
		str from = {0,0,0};
		str_copys(&from, sender);
		str domain = {0,0,0};
		str_copys(&domain, host);
		str_lower(&to);
		str_lower(&domain);
		str_lower(&from);
		do_forward(&to, &domain, &from);
		str_free(&to);
		str_free(&domain);
		str_free(&from);
	}
	if (do_deliver) {
		delivery_final();
		if (sender) {
			outmsgf("{Message from '}s{' to '}s{' delivered.\n}",
				sender, email);
		} else {
			outmsgf("{Message from 'null' to '}s{' delivered.\n}",
				email);
		}
	}
	if (sender) {
		autoreply();
	}
	if (run_under_vdeliver) {
		failfree();
		return SUCCESS_STOP;
	}
	failfree();
	return SUCCESS_CONTINUE;
}
