/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-  */
/* autoreply.c
 *
 * Copyright (C) 2019  Giam Teck Choon <choon@choon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <sys/wait.h>
#include <signal.h>
#include <bglibs/iobuf.h>
#include <bglibs/str.h>
#include "common.h"
#include "is_file.h"
#include "is_executable.h"

extern char *ext;
extern char *host;
extern char *sender;
extern char *recipient;
extern char *maildir;

extern str tmpstr;
extern str tmp2str;
extern str tmp3str;

extern void freemem(void);

static str mstr = {0,0,0};

static void failfree(void)
{
	if (mstr.s) str_free(&mstr);
	freemem();
}

/*
 * Return 1 if match otherwise return 0.
 */
static int str_filematchline(const char *filename, str *s)
{
	if (is_file(filename) && s->s) {
		str_lower(s);
		str_strip(s);
		ibuf cfn;
		if (!ibuf_open(&cfn, filename, 0)) {
			ibuf_close(&cfn);
			return 0;
		}
		static str msgstr = {0,0,0};
		while(ibuf_getstr(&cfn, &msgstr, LF)) {
			str_lower(&msgstr);
			str_strip(&msgstr);
			if (s->len >= msgstr.len && msgstr.len > 0) {
				if (strncmp(s->s, msgstr.s, msgstr.len) == 0) {
					errmsgf("{skipsender='}s{'\n}",
						msgstr.s);
					str_free(&msgstr);
					return 1;
				}
			}
		}
		str_free(&msgstr);
	}
	return 0;
}

void autoreply(void)
{
	if (!sender || sender[0] == 0) {
		return;
	}
	if (!is_executable("/usr/bin/qmail-autoresponder")) {
		return;
	}
	CATCH_NE_ONE(str_copys(&mstr, sender),
		     exit_cb(FAIL_SOFT, failfree));
	CATCH_NE_ONE(str_copy2s(&tmpstr, maildir, "/autoresponse/message.txt"),
		     exit_cb(FAIL_SOFT, failfree));
	CATCH_NE_ONE(str_copy2s(&tmp2str, maildir, "/autoresponse"),
		     exit_cb(FAIL_SOFT, failfree));
	CATCH_NE_ONE(str_copy2s(&tmp3str, maildir, "/autoresponse/.skipsender"),
		     exit_cb(FAIL_SOFT, failfree));
	if (!is_file(tmpstr.s)) {
		return;
	}
	if (str_filematchline(tmp3str.s, &mstr)) {
		return;
	}
	CATCH_NE_ZERO(lseek(STDIN_FILENO, 0L, SEEK_SET),
		      exit_cb(FAIL_SOFT, failfree));

	int fd1[2];
	CATCH_NE_ZERO(pipe(fd1), exit_cb(FAIL_SOFT, failfree));
	pid_t childpid;
	childpid = fork();
	if (childpid == -1) {
		errmsg("fork() failed\n");
		failfree();
		exit(FAIL_SOFT);
	} else if (childpid == 0) { /* child process */
		setenv("QMAILUSER", ext, 0);
		setenv("QMAILHOST", host, 0);
		close(fd1[0]);
		CATCH_NE(dup2(fd1[1], STDOUT_FILENO), STDOUT_FILENO,
			      exit_cb(FAIL_SOFT, failfree));
		close(fd1[1]);
		const char *cmd[] = {"/usr/bin/qmail-autoresponder",
			"-O", "no_inreplyto=1", tmp2str.s, 0};
		execvp(cmd[0], (char **)cmd);
		errmsg("execute /usr/bin/qmail-autoresponder failed\n");
		failfree();
		exit(FAIL_SOFT);
	} else { /* parent process */
		close(fd1[1]);
		signal(SIGPIPE, SIG_IGN);
		int status;
		if (waitpid(childpid, &status, WUNTRACED) != childpid) {
			errmsg("System call to 'waitpid' failed\n");
			close(fd1[0]);
			failfree();
			exit(FAIL_SOFT);
		}
		if (!WIFEXITED(status)) {
			errmsg("/usr/bin/qmail-autoresponder crashed\n");
			close(fd1[0]);
			failfree();
			exit(FAIL_SOFT);
		}
		if (WEXITSTATUS(status)) {
			errmsgf("{qmail-autoresponder exit status '}d{'\n}",
				WEXITSTATUS(status));
			static str msgstr = {0,0,0};
			ibuf r;
			ibuf_init(&r, fd1[0], 0, 0, 0);
			ibuf_seek(&r, 0);
			while(ibuf_getstr(&r, &msgstr, LF)) {
				errmsg(msgstr.s);
			}
			ibuf_close(&r);
			close(fd1[0]);
			str_free(&msgstr);
			exit(FAIL_SOFT);
		}
		close(fd1[0]);
	}
}
