/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-  */
/* check_spam.c
 *
 * Copyright (C) 2019  Giam Teck Choon <choon@choon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <stdint.h>
#include <errno.h>
#include <bglibs/str.h>
#include <bglibs/iobuf.h>
#include "common.h"
#include "is_file.h"
#include "mkmaildir.h"

extern char *home;
extern char *rpline;
extern char *dtline;
extern char *maildir;

extern str hostname;
extern str msgname;
extern str tmpstr;
extern str tmp2str;
extern str tmp3str;
extern uintmax_t msgsize;

extern void freemem(void);
extern void delivery_fail(void);
extern void delivery_final(void);

static struct stat st;
static str maildirtmp = {0,0,0};
static str userprefs = {0,0,0};
static str spamtmp = {0,0,0};
static str spamnew = {0,0,0};

static void failfree(void)
{
	if (spamtmp.s && is_file(spamtmp.s)) {
		unlink(spamtmp.s);
	}
	if (spamnew.s && is_file(spamnew.s)) {
		unlink(spamnew.s);
	}
	if (userprefs.s) str_free(&userprefs);
	if (spamtmp.s) str_free(&spamtmp);
	if (spamnew.s) str_free(&spamnew);
	if (maildirtmp.s) str_free(&maildirtmp);
	delivery_fail();
	freemem();
}

static void cleanup(void)
{
	if (spamtmp.s && maildirtmp.s && is_file(spamtmp.s)) {
		if (is_file(maildirtmp.s)) {
			CATCH_NE_ZERO(stat(spamtmp.s, &st),
				      exit_cb(FAIL_SOFT, failfree));
			uintmax_t spamsize = st.st_size;
			CATCH_NE_ZERO(stat(maildirtmp.s, &st),
				      exit_cb(FAIL_SOFT, failfree));
			uintmax_t tmpsize = st.st_size;
			if (spamsize > tmpsize) {
				CATCH_NE_ZERO(unlink(maildirtmp.s),
					      exit_cb(FAIL_SOFT, failfree));
				CATCH_NE_ZERO(link(spamtmp.s, maildirtmp.s),
					      exit_cb(FAIL_SOFT, failfree));
				CATCH_NE_ZERO(unlink(spamtmp.s),
					      exit_cb(FAIL_SOFT, failfree));
			} else {
				CATCH_NE_ZERO(unlink(spamtmp.s),
					      exit_cb(FAIL_SOFT, failfree));
			}
		} else {
			CATCH_NE_ZERO(link(spamtmp.s, maildirtmp.s),
				      exit_cb(FAIL_SOFT, failfree));
			CATCH_NE_ZERO(unlink(spamtmp.s),
				      exit_cb(FAIL_SOFT, failfree));
		}
	}
	if (spamtmp.s) str_free(&spamtmp);
	if (spamnew.s) str_free(&spamnew);
	if (maildirtmp.s) str_free(&maildirtmp);
}

/*
 * Return 1 for spam and 0 otherwise. 
 */
int check_spam(void)
{
	CATCH_NE_ONE(str_copy2s(&userprefs, maildir, SPAMVUSERPREF),
		     exit_cb(FAIL_SOFT, failfree));
	if (!is_file(userprefs.s)) {
		CATCH_NE_ONE(str_copy2s(&userprefs, home,
					"/.spamassassin/user_prefs"),
			     exit_cb(FAIL_SOFT, failfree));
	}
	if (!is_file(userprefs.s)) {
		if (userprefs.s) str_free(&userprefs);
		if (spamtmp.s) str_free(&spamtmp);
		if (spamnew.s) str_free(&spamnew);
		if (maildirtmp.s) str_free(&maildirtmp);
		return 0;
	}
	/* Create Spam sub maildir */
	str_truncate(&tmpstr, 0);
	CATCH_NE_ONE(str_copy3s(&tmpstr, maildir, "/.", SPAMDIRNAME),
		     exit_cb(FAIL_SOFT, failfree));
	CATCH_NE_ZERO(mkmaildir(tmpstr.s),
		      exit_cb(FAIL_SOFT, failfree));

	CATCH_NE_ONE(str_copy3s(&maildirtmp, maildir, "/tmp/", msgname.s),
		     exit_cb(FAIL_SOFT, failfree));

	CATCH_NE_ONE(str_copy5s(&spamtmp, maildir, "/.", SPAMDIRNAME,
				"/tmp/", msgname.s),
		     exit_cb(FAIL_SOFT, failfree));

	CATCH_NE_ONE(str_copy5s(&spamnew, maildir, "/.", SPAMDIRNAME,
				"/new/", msgname.s),
		     exit_cb(FAIL_SOFT, failfree));
	
	int cfd[2];
	CATCH_NE_ZERO(pipe(cfd), exit_cb(FAIL_SOFT, failfree));

	pid_t childpid;
	childpid = fork();
	if (childpid == -1) {
		errmsg("fork() failed\n");
		failfree();
		exit(FAIL_SOFT);
	} else if (childpid == 0) { /* child process */
		close(cfd[1]); /* close the fd of write end of the pipe */
		CATCH_NE(dup2(cfd[0], STDIN_FILENO), STDIN_FILENO,
			      exit_cb(FAIL_SOFT, failfree));
		close(cfd[0]); /* close the fd of read end of the pipe */
		int newfd = open(spamtmp.s, O_RDWR|O_CREAT|O_EXCL, 0600);
		if (newfd < 0) {
			errmsgf("{Error open/write file '}s{'\n}", spamtmp.s);
			failfree();
			exit(FAIL_SOFT);
		}
		CATCH_NE(dup2(newfd, STDOUT_FILENO), STDOUT_FILENO,
			      exit_cb(FAIL_SOFT, failfree));
		close(newfd);
		obuf_putsflush(&outbuf, rpline);
		obuf_putsflush(&outbuf, dtline);
		char *cmd[4];
		int opt = 0;
		cmd[opt] = "/usr/bin/spamassassin";
		cmd[opt++] = "spamassassin";
		cmd[opt++] = "-p";
		cmd[opt++] = userprefs.s;
		cmd[opt++] = "-e";
		cmd[opt] = (char *)NULL;
		execvp(cmd[0], cmd);
		errmsg("execute /usr/bin/spamassassin failed\n");
		failfree();
		exit(FAIL_SOFT);
	} else { /* parent process */
		close(cfd[0]); /* close the fd of the read end of pipe */
		signal(SIGPIPE, SIG_IGN);
		CATCH_NE_ZERO(lseek(STDIN_FILENO, 0L, SEEK_SET),
			      exit_cb(FAIL_SOFT, failfree));
		/*
		 * Copy content of STDIN_FILENO into cfd[1]
		 */
		ibuf i;
		ibuf_init(&i, STDIN_FILENO, 0, 0, 0);
		ibuf_copytofd(&i, cfd[1]);
		ibuf_close(&i);
		close(cfd[1]); /* close the fd of the write end of pipe */
		int status;
		if (waitpid(childpid, &status, WUNTRACED) != childpid) {
			errmsg("System call to 'waitpid' failed\n");
			failfree();
			exit(FAIL_SOFT);
		}
		if (!WIFEXITED(status)) {
			errmsg("child cmd spamassassin crashed\n");
			failfree();
			exit(FAIL_SOFT);
		}
		CATCH_NE_ZERO(lseek(STDIN_FILENO, 0L, SEEK_SET),
			      exit_cb(FAIL_SOFT, failfree));
		if (WEXITSTATUS(status)) {
			CATCH_NE_ZERO(stat(spamtmp.s, &st),
				      exit_cb(FAIL_SOFT, failfree));
			uintmax_t spamsize;
			if (st.st_blocks * 512 < st.st_size) {
				spamsize = st.st_size;
			} else {
				spamsize = st.st_blocks * 512;
			}
			if (spamsize < msgsize) {
				if (spamtmp.s && is_file(spamtmp.s)) {
					CATCH_NE_ZERO(unlink(spamtmp.s),
						      exit_cb(FAIL_SOFT,
							      failfree));
				}
				return 0;
			}
			CATCH_NE_ZERO(link(spamtmp.s, spamnew.s),
				      exit_cb(FAIL_SOFT, failfree));
			CATCH_NE_ZERO(unlink(spamtmp.s),
				      exit_cb(FAIL_SOFT, failfree));
			return 1;
		}
	}
	cleanup();
	return 0;
}
