/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-  */
/* forwarders.c
 *
 * Copyright (C) 2019  Giam Teck Choon <choon@choon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/wait.h>
#include <bglibs/iobuf.h>
#include <bglibs/str.h>
#include <bglibs/striter.h>
#include "common.h"

extern char *rpline;
extern char *dtline;
extern char *sender;
extern char *email;

extern str hostname;
extern str tmpstr;
extern str tmp2str;
extern str tmp3str;

extern void freemem(void);
extern void delivery_fail(void);

static str wenvstr = {0,0,0};

static void failfree(void)
{
	delivery_fail();
	if (wenvstr.s) str_free(&wenvstr);
	freemem();
}

static void write_envelope(str *o, str *from, str *to, str *domain)
{
	str fo = {0,0,0};
	str_catc(&fo, 'F');
	str_cats(&fo, from->s);
	str_catc(&fo, '\0');

	str v = {0,0,0};
	str_copyb(&v, to->s, to->len);
	str_lower(&v);
	str t = {0,0,0};
	striter iter;
	for (striter_start(&iter, &v, 0); striter_valid(&iter);
	     striter_advance(&iter)) {
		//str_truncate(&t, 0);
		str_copys(&t, (char*)iter.startptr);
		int at = str_findnext(&t, '@', 0);
		str_catc(&fo, 'T');
		if (at < 0) {
			// If the address has no '@', add the virtual domain
			str_cat2s(&t, "@", domain->s);
		} else if ((unsigned)at == t.len-1) {
			// If it has an '@', but no domain, add the local domain
			str_cats(&t, hostname.s);
		}
		str_cats(&fo, t.s);
		str_catc(&fo, '\0');
	}
	str_free(&v);
	str_free(&t);
	str_catc(&fo, '\0');
	str_copyb(o, fo.s, fo.len);
	str_free(&fo);
}

static void inject(str *from, str *to, str *domain)
{
	int pipe1[2];
	int pipe2[2];
	CATCH_NE_ZERO(pipe(pipe1), exit_cb(FAIL_SOFT, failfree));
	CATCH_NE_ZERO(pipe(pipe2), exit_cb(FAIL_SOFT, failfree));
	pid_t childpid;
	childpid = fork();
	if (childpid == -1) {
		errmsg("System call to 'fork' failed.\n");
		failfree();
		exit(FAIL_SOFT);
	} else if (childpid == 0) {
		close(pipe1[1]);
		close(pipe2[1]);
		CATCH_NE(dup2(pipe1[0], STDIN_FILENO), STDIN_FILENO,
			 exit_cb(FAIL_SOFT, failfree));
		CATCH_NE(dup2(pipe2[0], STDOUT_FILENO), STDOUT_FILENO,
			 exit_cb(FAIL_SOFT, failfree));
		execl("/var/qmail/bin/qmail-queue", "qmail-queue", (char *)NULL);
 		errmsg("Exec of qmail-queue failed.\n");
		failfree();
		exit(FAIL_SOFT);
	} else {
		close(pipe1[0]);
		close(pipe2[0]);
		signal(SIGPIPE, SIG_IGN);

		CATCH_NE(lseek(STDIN_FILENO, 0L, SEEK_SET), 0,
			 exit_cb(FAIL_SOFT, failfree));

		obuf op1;
		obuf_init(&op1, pipe1[1], 0, 0, 0);
		obuf_putsflush(&op1, rpline);
		obuf_putsflush(&op1, dtline);
		//obuf_putsflush(&op1, "X-Forwarded-From: ");
		//obuf_putsflush(&op1, email);
		//obuf_putsflush(&op1, CRLF);
		obuf_copyfromfd(STDIN_FILENO, &op1);
		obuf_sync(&op1);
		obuf_close(&op1);
		close(pipe1[1]);

		write_envelope(&wenvstr, from, to, domain);
		//str_subst(&wenvstr, '|', '\0');
		obuf op2;
		obuf_init(&op2, pipe2[1], 0, 0, 0);
		unsigned int j;
		for (j=0; j<wenvstr.len; j++) {
			obuf_putc(&op2, wenvstr.s[j]);
			obuf_flush(&op2);
		}
		obuf_sync(&op2);
		obuf_close(&op2);
		close(pipe2[1]);

		int status;
		if (waitpid(childpid, &status, WUNTRACED) != childpid) {
	 		errmsg("System call to 'waitpid' failed.\n");
			failfree();
			exit(FAIL_SOFT);
		}
		if (!WIFEXITED(status)) {
	 		errmsg("qmail-queue crashed!\n");
			failfree();
			exit(FAIL_SOFT);
		}
		if (WEXITSTATUS(status)) {
			errmsgf("{status='}d{'\n}", WEXITSTATUS(status));
	 		errmsg("qmail-queue exited with an error!\n");
			failfree();
			exit(FAIL_SOFT);
		}
		int x = str_findnext(&wenvstr, 'T', 0);
		str_lcut(&wenvstr, x+1);
		str_rstrip(&wenvstr);
		str_subst(&wenvstr, '\0', ',');
		str_subst(&wenvstr, 'T', ' ');
		str_truncate(&wenvstr, wenvstr.len-2);
		outmsgf("{Message from '}s{' to '}s{' forward to '}s{'\n}",
			sender, email, wenvstr.s);
		str_free(&wenvstr);
	}
}

void do_forward(str *to, str *domain, str *from)
{
	int at = str_findlast(from, '@');
	if (at > 0) {
		str_truncate(&tmpstr, 0);
		CATCH_NE_ONE(str_copyb(&tmpstr, from->s, at),
			     exit_cb(FAIL_SOFT, failfree));
		CATCH_NE_ZERO(setenv("QMAILUSER", tmpstr.s, 0),
			      exit_cb(FAIL_SOFT, failfree));
		str_truncate(&tmpstr, 0);
		at++;
		CATCH_NE_ONE(str_copyb(&tmpstr, from->s+at, from->len-at),
			     exit_cb(FAIL_SOFT, failfree));
		CATCH_NE_ZERO(setenv("QMAILHOST", tmpstr.s, 0),
			      exit_cb(FAIL_SOFT, failfree));
	}
	str_rstrip(from);
	inject(from, to, domain);
}
