/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-  */
/* vquota.c
 *
 * Copyright (C) 2019  Giam Teck Choon <choon@choon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <stdint.h>
#include <limits.h>
#include <bglibs/str.h>
#include <bglibs/vmailmgr/vpwentry.h>
#include "common.h"
#include "is_maildir.h"
#include "is_dir.h"

extern char *maildir;

extern str hostname;
extern str msgname;
extern str tmpstr;
extern str tmp2str;
extern str tmp3str;

extern uintmax_t msgsize;

extern void freemem(void);

static struct stat st;


static void wrap_stat(const char *path)
{
	if (stat(path, &st) == -1) {
		errmsgf("{stat '}s{' failed\n}", path);
		freemem();
		exit(FAIL_SOFT);
	}
	//CATCH_NE_ZERO(stat(path, &st), exit_cb(FAIL_SOFT, freemem));
}

static uintmax_t stat_size(const char *path)
{
	wrap_stat(path);
	if (st.st_blocks * 512 < st.st_size) {
		return st.st_size;
	} else {
		return st.st_blocks * 512;
	}
}

static void scan_dir(const char *path,
		     uintmax_t *count, uintmax_t *size,
		     void (*fn)(const char*, const str*,
				uintmax_t*, uintmax_t*))
{
	str fullname = {0,0,0};
	DIR *dir;
	direntry *entry;

	if ((dir = opendir(path)) == 0) { /* <sys/types.h> and <dirent.h> */
		errmsgf("{Could not open directory '}s{'\n}", path);
		freemem();
		exit(FAIL_SOFT);
	}

	while ((entry = readdir(dir)) != 0) {
		const char *name = entry->d_name;
		if (name[0] == '.' && (name[1] == 0
				       || (name[1] == '.'
					   && name[2] == 0))) {
			continue;
		}
		CATCH_NE_ONE(str_copy3s(&fullname, path, "/", name),
			     exit_cb(FAIL_SOFT, freemem));
		fn(name, &fullname, count, size);
	}
	closedir(dir);
	str_free(&fullname);
}

static void stat_file(const char *name, const str *path,
		      uintmax_t *count, uintmax_t *size)
{
	wrap_stat(path->s);
	if (S_ISREG(st.st_mode)) {
		++*count;
		*size += st.st_blocks * 512;
	}
	/* Shut up warnings */
	name = 0;
}

/*
 * Scan the cur/new subdirectories of a maildir.
 */
static void stat_maildir(const str *path,
			 uintmax_t *count, uintmax_t *size)
{
	str tmp = {0,0,0};
	if (!str_ready(&tmp, path->len + 4)) {
		errno = ENOMEM;
		errmsgf("{File='}s{' Function='}s{' Line='}d{'"
			" Errno='}d{' Errmsg='}s{'\n}",
			__FILE__, __FUNCTION__, __LINE__,
			errno, strerror(errno));
		freemem();
		exit(FAIL_SOFT);
	}
	str_copy(&tmp, path);
	str_cats(&tmp, "/cur");
	if (is_dir(tmp.s)) {
		*size += stat_size(tmp.s);
		scan_dir(tmp.s, count, size, stat_file);
	}

	str_copy(&tmp, path);
	str_cats(&tmp, "/new");
	if (is_dir(tmp.s)) {
		*size += stat_size(tmp.s);
		scan_dir(tmp.s, count, size, stat_file);
	}

	str_copy(&tmp, path);
	str_cats(&tmp, "/tmp");
	if (is_dir(tmp.s)) {
		*size += stat_size(tmp.s);
		scan_dir(tmp.s, count, size, stat_file);
	}

	str_free(&tmp);
}


/*
 * Stat a file (don't count it), or recurse into maildirs.
 */
static void stat_toplevel(const char *entry, const str *path,
			  uintmax_t *count, uintmax_t *size)
{
	wrap_stat(path->s);
	*size += st.st_blocks * 512;
	if (S_ISDIR(st.st_mode)) {
		if (strcmp(entry, "cur") == 0 || strcmp(entry, "new") == 0) {
			scan_dir(path->s, count, size, stat_file);
		}
		else if (strcmp(entry, "tmp") != 0) {
			stat_maildir(path, count, size);
		}
	}
}

static void size_mailbox(const char *path,
			 uintmax_t *count, uintmax_t *size)
{
	*size = stat_size(path);
	*count = 0;
	scan_dir(path, count, size, stat_toplevel);
}

/*
 * Ref: https://github.com/bruceg/vmailmgr-tools/blob/master/vcheckquota.c
 */
void check_quota(vpwentry *vpw)
{
	if (!is_maildir(maildir)) {
		return;
	}
	/*
	 * There are 4 cases to consider when comparing disk usage (du)
	 * agains hard and soft quotas:
	 *
	 * Case 1: soft = 0, hard = 0: user has no quota set
	 * Case 2: soft = 0, hard > 0: use hard quota
	 * Case 3: soft > 0, hard = 0: treat soft quota as hard
	 * Case 4: soft > 0, hard > 0: if du becomes larger
	 *         then soft quota, allow message in if
	 *         a) it is small (<2048 bytes),
	 *         b) it would not put du over hard quota.
	 */
	static uintmax_t count = 0;
	static uintmax_t size = 0;
	static uintmax_t hardquota;
	static uintmax_t softquota;

	hardquota = vpw->hardquota;
	softquota = vpw->softquota;
	if (vpw->msgsize != UINT_MAX && msgsize > vpw->msgsize) {
		errmsg("Sorry, this message is larger than the current "
		       "maximum message size limit.\n");
		freemem();
		exit(FAIL_SOFT);
	}
	/*
	 * Case 1: no quotas set.
	 */
	if (vpw->softquota == UINT_MAX && vpw->hardquota == UINT_MAX
		&& vpw->msgcount == UINT_MAX) {
		return;
	}
	size_mailbox(maildir, &count, &size);
	size += msgsize;
	++count;
	/*
	 * Too many messages in the mailbox.
	 */
	if (vpw->msgcount != UINT_MAX && count > vpw->msgcount) {
    		errmsg("Sorry, the person you sent this message has too many"
		       " messages stored in the mailbox\n");
		freemem();
		exit(FAIL_SOFT);
	}
	/*
	 * No total size quotas are set.
	 */
	if (hardquota == UINT_MAX && softquota == UINT_MAX) {
		return;
	}
	/*
	 * Take care of Cases 2 and 3, and make everything look like Case 4.
	 */
	else {
		if (softquota > hardquota) {
			hardquota = softquota;
		}
		if (hardquota != UINT_MAX) {
			/* Convert B to MB */
			hardquota = hardquota * 1024 * 1024;
		}
	}
	/*
	 * Check hard quota before soft quota, as it has priority.
	 */
	if (size > hardquota) {
		errmsg("Message would exceed virtual user's disk quota.\n");
		freemem();
		exit(FAIL_SOFT);
	}
}
